import numpy as np
from scipy import stats
from copy import deepcopy

try:
    import cupy as cp

except (ModuleNotFoundError, ImportError) as e:
    pass


class UniformDistribution(object):
    """Generate uniform distribution between ``min`` and ``max``

    Args:
        min_val (double): Minimum in the uniform distribution
        max_val (double): Maximum in the uniform distribution
        use_cupy (bool, optional): If ``True``, use CuPy. If ``False`` use Numpy.
            (default: ``False``)

    Raises:
        ValueError: Issue with inputs. 

    """

    def __init__(self, min_val, max_val, use_cupy=False):

        if min_val > max_val:
            tmp = min_val
            min_val = max_val
            max_val = tmp
        elif min_val == max_val:
            raise ValueError("Min and max values are the same.")

        self.min_val = min_val
        self.max_val = max_val
        self.diff = max_val - min_val

        self.pdf_val = 1 / self.diff
        self.logpdf_val = np.log(self.pdf_val)

        self.use_cupy = use_cupy
        if use_cupy:
            try:
                cp.abs(1.0)
            except NameError:
                raise ValueError("CuPy not found.")

    def rvs(self, size=1):

        if not isinstance(size, int) and not isinstance(size, tuple):
            raise ValueError("size must be an integer or tuple of ints.")

        if isinstance(size, int):
            size = (size,)

        xp = np if not self.use_cupy else cp

        rand_unif = xp.random.rand(*size)

        out = rand_unif * self.diff + self.min_val

        return out

    def pdf(self, x):

        out = self.pdf_val * ((x >= self.min_val) & (x <= self.max_val))

        return out

    def logpdf(self, x):

        xp = np if not self.use_cupy else cp

        out = xp.zeros_like(x)
        out[(x >= self.min_val) & (x <= self.max_val)] = self.logpdf_val
        out[(x < self.min_val) | (x > self.max_val)] = -np.inf
        return out

    def copy(self):
        breakpoint()
        return deepcopy(self)


def uniform_dist(min, max, use_cupy=False):
    """Generate uniform distribution between ``min`` and ``max``

    Args:
        min (double): Minimum in the uniform distribution
        max (double): Maximum in the uniform distribution
        use_cupy (bool, optional): If ``True``, use CuPy. If ``False`` use Numpy.
            (default: ``False``)

    Returns:
        :class:`UniformDistribution`: Uniform distribution.


    """
    dist = UniformDistribution(min, max, use_cupy=use_cupy)

    return dist


def log_uniform(min, max):
    """Generate log-uniform distribution between ``min`` and ``max``

    Args:
        min (double): Minimum in the log-uniform distribution
        max (double): Maximum in the log-uniform distribution

    Returns:
        scipy distribution object: Log-uniform distribution built from
            `scipy.stats.uniform <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.loguniform.html>_`.

    """
    # adjust ordering if needed
    if min > max:
        temp = min
        min = max
        max = temp

    # setup quantities for scipy
    sig = max - min
    dist = stats.loguniform(min, sig)
    return dist


class MappedUniformDistribution:
    """Maps uniform distribution to zero to 1.

    This is a modified uniform distribution that maps
    the input values to a range from zero to 1 by using ``min`` and
    ``max`` values input by user. This ensures the log of the prior value
    from this distribution is zero if the value is between ``min`` and ``max``.
    and ``-np.inf`` if it is outside that range.

    Args:
        min (double): Minimum in the uniform distribution
        max (double): Maximum in the uniform distribution
        use_cupy (bool, optional): If ``True``, use CuPy. If ``False`` use Numpy.
            (default: ``False``)

    Raises:
        ValueError: If ``min`` is greater than ``max``.


    """

    def __init__(self, min, max, use_cupy=False):
        self.min, self.max = min, max
        self.diff = self.max - self.min
        if self.min > self.max:
            raise ValueError("min must be less than max.")

        self.dist = uniform_dist(0.0, 1.0, use_cupy=use_cupy)

    def logpdf(self, x):
        """Get the log of the pdf value for this distribution.

        Args:
            x (double np.ndarray):
                Input parameters to get prior values.

        Returns:
            np.ndarray: Associated logpdf values of the input.

        """
        temp = 1.0 - (self.max - x) / self.diff
        return self.dist.logpdf(temp)

    def rvs(self, size=1):
        """Get the log of the pdf value for this distribution.

        Args:
            size (int or tuple of ints, optional): Output size for number of generated
                sources from prior distributions.

        Returns:
            np.ndarray: Generated values.

        """
        # adjust size if int
        if isinstance(size, int):
            size = (size,)

        elif not isinstance(size, tuple):
            raise ValueError("Size must be int or tuple of ints.")

        temp = self.dist.rvs(size=size)

        return self.max + (temp - 1.0) * self.diff


class ProbDistContainer:
    """Container for holding and generating prior info

    Args:
        priors_in (dict): Dictionary with keys as int or tuple of int
            describing which parameters the prior takes. Values are
            probability distributions with ``logpdf`` and ``rvs`` methods.

    Attributes:
        priors_in (dict): Dictionary with keys as int or tuple of int
            describing which parameters the prior takes. Values are
            probability distributions with ``logpdf`` and ``rvs`` methods.
        priors (list): list of indexes and their associated distributions arranged
            in a list.
        ndim (int): Full dimensionality.
        use_cupy (bool, optional): If ``True``, use CuPy. If ``False`` use Numpy.
            (default: ``False``)

    Raises:
        ValueError: Missing parameters or incorrect index keys.

    """

    def __init__(self, priors_in, use_cupy=False):

        # copy to have
        self.priors_in = priors_in.copy()

        # to separate out in list form
        self.priors = []

        # setup lists
        temp_inds = []
        for inds, dist in priors_in.items():

            # multiple index
            if isinstance(inds, tuple):
                inds_in = np.asarray(inds)
                self.priors.append([inds_in, dist])

            # single index
            elif isinstance(inds, int):
                inds_in = np.array([inds])
                self.priors.append([inds_in, dist])

            else:
                raise ValueError(
                    "Keys for prior dictionary must be an integer or tuple."
                )

            temp_inds.append(np.asarray([inds_in]))

        uni_inds = np.unique(np.concatenate(temp_inds, axis=1).flatten())
        if len(uni_inds) != len(np.arange(np.max(uni_inds) + 1)):
            raise ValueError(
                "Please ensure all sampled parameters are included in priors."
            )

        self.ndim = uni_inds.max() + 1

        self.use_cupy = use_cupy

        for key, item in self.priors_in.items():
            item.use_cupy = use_cupy

    def logpdf(self, x):
        """Get logpdf by summing logpdf of individual distributions

        Args:
            x (double np.ndarray[..., ndim]):
                Input parameters to get prior values.

        Returns:
            np.ndarray[...]: Prior values.

        """
        # TODO: check if mutliple index prior will work
        xp = np if not self.use_cupy else cp

        # make sure at least 2D
        if x.ndim == 1:
            x = x[None, :]
            squeeze = True

        elif x.ndim != 2:
            raise ValueError("x needs to 1 or 2 dimensional array.")
        else:
            squeeze = False

        prior_vals = xp.zeros(x.shape[0])

        # sum the logs (assumes parameters are independent)
        for i, (inds, prior_i) in enumerate(self.priors):
            vals_in = x[:, inds].squeeze()
            if hasattr(prior_i, "logpdf"):
                temp = prior_i.logpdf(vals_in)
            else:
                temp = prior_i.logpmf(vals_in)

            prior_vals += temp

        # if only one walker was asked for, return a scalar value not an array
        if squeeze:
            prior_vals = prior_vals[0].item()

        return prior_vals

    def ppf(self, x, groups=None):
        """Get logpdf by summing logpdf of individual distributions

        Args:
            x (double np.ndarray[..., ndim]):
                Input parameters to get prior values.

        Returns:
            np.ndarray[...]: Prior values.

        """

        if groups is not None:
            raise NotImplementedError

        xp = np if not self.use_cupy else cp

        # TODO: check if mutliple index prior will work
        is_1d = x.ndim == 1
        x = xp.atleast_2d(x)
        out_vals = xp.zeros_like(x)

        # sum the logs (assumes parameters are independent)
        for i, (inds, prior_i) in enumerate(self.priors):
            if len(inds) > 1:
                raise NotImplementedError

            vals_in = x[:, inds].squeeze()
            temp = prior_i.ppf(vals_in)

            out_vals[:, inds[0]] = temp

        if is_1d:
            return out_vals.squeeze()
        return out_vals

    def rvs(self, size=1):
        """Generate random values according to prior distribution

        The user will have to be careful if there are prior functions that
        do not have an ``rvs`` method. This means that generated points may lay
        inside the prior of all input priors that have ``rvs`` methods, but
        outside the prior if priors without the ``rvs`` method are included.

        Args:
            size (int or tuple of ints, optional): Output size for number of generated
                sources from prior distributions.

        Returns:
            np.ndarray[``size + (self.ndim,)``]: Generated samples.

        Raises:
            ValueError: If size is not an int or tuple.


        """

        # adjust size if int
        if isinstance(size, int):
            size = (size,)

        elif not isinstance(size, tuple):
            raise ValueError("Size must be int or tuple of ints.")

        xp = np if not self.use_cupy else cp

        # setup the slicing to properly sample points
        out_inds = tuple([slice(None) for _ in range(len(size))])

        # setup output and loop through priors
        out = xp.zeros(size + (self.ndim,))
        for i, (inds, prior_i) in enumerate(self.priors):
            # guard against extra prior functions without rvs methods
            if not hasattr(prior_i, "rvs"):
                continue
            # combines outer dimensions with indices of interest
            inds_in = out_inds + (inds,)

            # allows for proper adding of quantities to out array
            if len(inds) == 1:
                adjust_inds = out_inds + (None,)
                out[inds_in] = prior_i.rvs(size=size)[adjust_inds]
            else:
                out[inds_in] = prior_i.rvs(size=size)

        return out



# -*- coding: utf-8 -*-

import numpy as np
from .mh import MHMove
from ..prior import ProbDistContainer

__all__ = ["DistributionGenerate"]


class DistributionGenerate(MHMove):
    """Generate proposals from a distribution

    Args:
        generate_dist (object): :class:`ProbDistContainer` object that has ``logpdf``
            and ``rvs`` methods.

    """

    def __init__(self, generate_dist, *args, **kwargs):

        for key in generate_dist:
            if not isinstance(generate_dist[key], ProbDistContainer):
                raise ValueError(
                    "Distributions need to be eryn.prior.ProbDistContainer object."
                )
        self.generate_dist = generate_dist
        super(DistributionGenerate, self).__init__(*args, **kwargs)

    def get_proposal(self, all_coords, all_inds, random, **kwargs):
        """Make a proposal

        Args:
            all_coords (dict): Keys are ``branch_names``. Values are
                np.ndarray[ntemps, nwalkers, nleaves_max, ndim]. These are the curent
                coordinates for all the walkers.
            all_inds (dict): Keys are ``branch_names``. Values are
                np.ndarray[ntemps, nwalkers, nleaves_max]. These are the boolean
                arrays marking which leaves are currently used within each walker.

        Returns:
            tuple: Tuple containing proposal information.
                First entry is the new coordinates as a dictionary with keys
                as ``branch_names`` and values as
                ``double `` np.ndarray[ntemps, nwalkers, nleaves_max, ndim] containing
                proposed coordinates. Second entry
                is the factors associated with the
                proposal necessary for detailed balance. This is effectively
                any term in the detailed balance fraction. +log of factors if
                in the numerator. -log of factors if in the denominator.

        """

        # set up all initial holders
        q = {}
        factors = {}
        new_inds = {}
        if all_inds is None:
            all_inds = {
                name: np.ones(coords.shape[:-1], dtype=bool)
                for name, coords in all_coords
            }

        # iterate through branches and propose new points where inds == True
        for i, (name, coords, inds) in enumerate(
            zip(all_coords.keys(), all_coords.values(), all_inds.values(),)
        ):
            # copy over previous info
            ntemps, nwalkers, _, _ = coords.shape
            q[name] = coords.copy()
            new_inds[name] = inds.copy()

            if i == 0:
                factors = np.zeros((ntemps, nwalkers))

            # add coordinates for new leaves
            current_generate_dist = self.generate_dist[name]
            inds_here = np.where(inds == True)
            num_inds_change = len(inds_here[0])

            old_points = coords[inds_here]

            # old points so + log(qold)
            factors[inds_here[:2]] += +1 * current_generate_dist.logpdf(old_points)

            # Draw
            new_points = current_generate_dist.rvs(size=num_inds_change)

            # new point, so -log(qnew)
            factors[inds_here[:2]] += -1 * current_generate_dist.logpdf(new_points)

            q[name][inds_here] = new_points

        return q, factors



# -*- coding: utf-8 -*-
try:
    import cupy as xp
except (ModuleNotFoundError, ImportError):
    pass

import numpy as np

from .group import GroupMove
from .stretch import StretchMove

__all__ = ["GroupStretchMove"]


class GroupStretchMove(GroupMove, StretchMove):
    """Proposal like stretch with stationary compliment.

    This move uses the stretch proposal method and math, but the compliment
    of walkers used to propose a new point is chosen from a stationary group
    rather than the current walkers in the ensemble. 

    This move allows for "stretch"-like proposal to be used in Reversible Jump MCMC.

    Args:
        **kwargs (dict, optional): Keyword arguments passed to :class:`GroupMove` and 
            :class:`StretchMove`.

    """

    def __init__(self, **kwargs):
        GroupMove.__init__(self, **kwargs)
        StretchMove.__init__(self, **kwargs)

    def get_proposal(self, s_all, random, gibbs_ndim=None, s_inds_all=None, **kwargs):
        """Generate group stretch proposal coordinates

        Args:
            s_all (dict): Keys are ``branch_names`` and values are coordinates
                for which a proposal is to be generated.
            random (object): Random state object.
            gibbs_ndim (int or np.ndarray, optional): If Gibbs sampling, this indicates
                the true dimension. If given as an array, must have shape ``(ntemps, nwalkers)``.
                See the tutorial for more information.
                (default: ``None``)
            s_inds_all (dict, optional): Keys are ``branch_names`` and values are 
                ``inds`` arrays indicating which leaves are currently used. (default: ``None``)

        Returns:
            tuple: First entry is new positions. Second entry is detailed balance factors.

        Raises:
            ValueError: Issues with dimensionality.

        """
        # needs to be set before we reach the end
        self.zz = None
        random_number_generator = random if not self.use_gpu else self.xp.random
        newpos = {}

        # iterate over branches
        for i, name in enumerate(s_all):
            # get points to move
            s = self.xp.asarray(s_all[name])

            Ns = s.shape[1]

            if s_inds_all is not None:
                s_inds = self.xp.asarray(s_inds_all[name])
            else:
                s_inds = None

            ntemps, nwalkers, nleaves_max, ndim_here = s.shape

            # gets rid of any values of exactly zero
            ndim_temp = nleaves_max * ndim_here

            # need to properly handle ndim
            if i == 0:
                ndim = ndim_temp
                Ns_check = Ns

            else:
                ndim += ndim_temp
                if Ns_check != Ns:
                    raise ValueError("Different number of walkers across models.")


            Ns = nwalkers

            # get actual compliment values
            c_temp = self.choose_c_vals(name, s, s_inds=s_inds)

            # use stretch to get new proposals
            newpos[name] = self.get_new_points(
                name, s, c_temp, Ns, s.shape, i, random_number_generator
            )

        # proper factors
        factors = (ndim - 1.0) * self.xp.log(self.zz)
        if self.use_gpu and not self.return_gpu:
            factors = factors.get()

        if gibbs_ndim is not None:
            # adjust factors in place
            self.adjust_factors(factors, ndim, gibbs_ndim)

        return newpos, factors




# -*- coding: utf-8 -*-

from copy import deepcopy

try:
    import cupy as xp

except (ModuleNotFoundError, ImportError) as e:
    import numpy as np

import numpy as np

__all__ = ["State"]


class BranchSupplimental(object):
    """Special object to carry information through sampler.

    The :class:`BranchSupplimental` object is a holder of information that is
    passed through the sampler. It can also be indexed similar to other quantities
    carried throughout the sampler.

    This indexing is based on the ``base_shape``. You can store many objects that have the same base
    shape and then index across all of them. For example, if you want to store individual leaf 
    information, the base shape will be ``(ntemps, nwalkers, nleaves_max)``. 
    If you want to store a 2D array per individual leaf, the overall shape will be 
    ``(ntemps, nwalkers, nleaves_max, dim2_extra, dim1_extra)``. Another type of information is
    stored in a class object (for example). Using ``numpy`` object arrays, 
    ``ntemps * nwalkers * nleaves_max`` number of class objects can be stored in the array. Then,
    using special indexing functions, information can be updated/accessed across all objects 
    stored simultaneously. If you index this class, it will give you back a dictionary with
    all objects stored indexed for each leaf. So if you index (0, 0, 0) in our running example,
    you will get back a dictionary with one 2D array and one class object from the ``numpy`` object
    array. 

    All of these objects are stored in ``self.holder``.

    Args:
        obj_info (dict): Initial information for storage. Keys are the names to be stored under
            and values are arrays. These arrays should have a base shape that is equivalent to
            ``base_shape``, meaning ``array.shape[:len(base_shape)] == self.base_shape``. 
            The dimensions beyond the base shape can be anything.
        base_shape (tuple): Base shape for indexing. Objects stored in the supplimental object 
            will have a shape that at minimum is equivalent to ``base_shape``.
        copy (bool, optional): If ``True``, copy whatever information is given in before it is stored. 
            if ``False``, store directly the input information. (default: ``False``) 

    Attributes:
        holder (dict): All of the objects stored for this supplimental object.

    
    """

    def __init__(self, obj_info: dict, base_shape: tuple, copy: bool = False):

        # store initial information
        self.holder = {}
        self.base_shape = base_shape
        self.ndim = len(self.base_shape)

        # add initial set of objects
        self.add_objects(obj_info, copy=copy)

    def add_objects(self, obj_info: dict, copy=False):
        """Add objects to the holder.
        
        Args:
            obj_info (dict): Information for storage. Keys are the names to be stored under
                and values are arrays. These arrays should have a base shape that is equivalent to
                ``base_shape``, meaning ``array.shape[:len(base_shape)] == self.base_shape``. 
                The dimensions beyond the base shape can be anything.
            copy (bool, optional): If ``True``, copy whatever information is given in before it is stored. 
                if ``False``, store directly the input information. (default: ``False``) 

        Raises:
            ValueError: Shape matching issues.
        
        """

        # whether a copy is requested
        dc = deepcopy if copy else (lambda x: x)

        # iterate through the dictionary of incoming objects to add
        for name, obj_contained in obj_info.items():
            if (
                isinstance(obj_contained, np.ndarray)
                and obj_contained.dtype.name == "object"
            ):
                self.holder[name] = dc(obj_contained)
                if self.base_shape is None:
                    self.base_shape = self.holder[name].shape
                    self.ndim = ndim = len(self.base_shape)
                else:
                    if self.holder[name].shape != self.base_shape:
                        raise ValueError(
                            f"Outer shapes of all input objects must be the same. {name} object array has shape {self.holder[name].shape}. The original shape found was {self.base_shape}."
                        )

            else:
                self.ndim = ndim = len(self.base_shape)

                # xp for GPU
                if isinstance(obj_contained, np.ndarray) or isinstance(
                    obj_contained, xp.ndarray
                ):
                    self.holder[name] = obj_contained.copy()

                # fill object array from list
                # adjust based on how many dimensions found
                else:
                    # objects to be stored
                    self.holder[name] = np.empty(self.base_shape, dtype=object)
                    if len(obj_contained) != self.base_shape[0]:
                        raise ValueError(
                            "Shapes of obj_contained does not match base_shape along axis 0."
                        )

                    if ndim > 1:
                        for i in range(self.base_shape[0]):
                            if len(obj_contained[i]) != self.base_shape[1]:
                                raise ValueError(
                                    "Shapes of obj_contained does not match obj_contained_sha along axis 1."
                                )

                            if ndim > 2:
                                for j in range(self.base_shape[1]):
                                    if len(obj_contained[i][j]) != self.base_shape[2]:
                                        raise ValueError(
                                            "Shapes of obj_contained does not match base_shape along axis 2."
                                        )

                                    for k in range(self.base_shape[2]):
                                        self.holder[name][i, j, k] = obj_contained[i][
                                            j
                                        ][k]
                            else:
                                for j in range(self.base_shape[1]):
                                    self.holder[name][i, j] = obj_contained[i][j]

                    else:
                        for i in range(self.base_shape[0]):
                            self.holder[name][i] = obj_contained[i]

    def remove_objects(self, names):
        """Remove objects from the holder.


        Args:
            names (str or list of str): Strings associated with information to delete.
                Please note it does not return the information. 

        Raises:
            ValueError: Input issues. 


        """
        # check inputs
        if not isinstance(names, list):
            if not isinstance(names, str):
                raise ValueError("names must be a string or list of strings.")

            names = [names]

        # iterate and remove items from holder
        for name in names:
            self.holder.pop(name)

    @property
    def contained_objects(self):
        """The list of keys of contained objects."""
        return list(self.holder.keys())

    def __contains__(self, name: str):
        """Check if the holder holds a specific key."""
        return name in self.holder

    def __getitem__(self, tmp):
        """Special indexing for retrieval.
        
        When indexing the overall class, this will return the slice of each object
        
        Args:
            tmp (int, np.ndarray, or slice): indexing slice of some form.

        Returns:
            dict: Keys are names of the objects contained. Values are the slices of those objects.
        
        """
        # slice each object contained
        return {name: values[tmp] for name, values in self.holder.items()}

    def __setitem__(self, tmp, new_value):
        """Special indexing for setting elements.
        
        When indexing the overall class, this will set object information.

        **Please note**: If you try to input information that is not already stored,
        it will ignore it.
        
        Args:
            tmp (int, np.ndarray, or slice): indexing slice of some form.

        """
        # loop through values already in holder
        for name, values in self.holder.items():

            if name not in new_value:
                continue
            # if the name is already contained, update with incoming value
            self.holder[name][tmp] = new_value[name]

    def take_along_axis(self, indices, axis: int, skip_names=[]):
        """Take information from contained arrays along an axis.
        
        See ```numpy.take_along_axis`` <https://numpy.org/doc/stable/reference/generated/numpy.take_along_axis.html>`_. 

        Args:
            indices (xp.ndarray): Indices to take along each 1d slice of arr. This must match the dimension 
                of ``self.base_shape``, but other dimensions only need to broadcast against 
                ``self.base_shape``.
            axis (int): The axis to take 1d slices along.
            skip_names (list of str, optional): By default, this function returns the results for 
                all stored objects. This list gives the strings of objects to leave behind and
                not return.

        Returns:
            dict: Keys are names of stored objects and values are the proper array slices.
        

        """
        # prepare output dictionary
        out = {}

        # iterate through holder
        for name, values in self.holder.items():
            # skip names if desired
            if name in skip_names:
                continue

            indices_temp = indices.copy()
            # adjust indices properly for specific object within the holder
            if (
                isinstance(values, np.ndarray) and values.dtype.name != "object"
            ) or isinstance(values, xp.ndarray):
                # expand the dimensions of the indexing values for non-object arrays
                for _ in range(values.ndim - indices_temp.ndim):
                    if isinstance(values, np.ndarray):
                        indices_temp = np.expand_dims(np.asarray(indices_temp), (-1,))
                    elif isinstance(values, xp.ndarray):
                        indices_temp = xp.expand_dims(xp.asarray(indices_temp), (-1,))

            # store the output for either numpy or cupy
            if isinstance(values, np.ndarray):
                out[name] = np.take_along_axis(values, indices_temp, axis)

            elif isinstance(values, xp.ndarray):
                out[name] = xp.take_along_axis(values, indices_temp, axis)

        return out

    def put_along_axis(self, indices, values_in: dict, axis: int):
        """Put information information into contained arrays along an axis.
        
        See ```numpy.put_along_axis`` <https://numpy.org/doc/stable/reference/generated/numpy.put_along_axis.html>`_. 

        **Please note** this function is not implemented in ``cupy``, so this is a custom implementation
        for both ``cupy`` and ``numpy``.

        Args:
            indices (xp.ndarray): Indices to put values along each 1d slice of arr. This must match 
            the dimension of ``self.base_shape``, but other dimensions only need to broadcast against 
            ``self.base_shape``.
            axis (int): The axis to put 1d slices along.
            values_in (dict): Keys are the objects contained to update. Values are the arrays of these
                objects with shape and dimension that can broadcast to match that of indices.

        """
        # iterate through all objects in the holder
        for name, values in self.holder.items():
            # skip names that are not to be updated
            if name not in values_in:
                continue

            # will need to have flexibility to broadcast
            indices_temp = indices.copy()

            if (
                isinstance(values, np.ndarray) and values.dtype.name != "object"
            ) or isinstance(values, xp.ndarray):
                # prepare indices for proper broadcasting
                for _ in range(values.ndim - indices_temp.ndim):
                    if isinstance(values, np.ndarray):
                        indices_temp = np.expand_dims(np.asarray(indices_temp), (-1,))
                    elif isinstance(values, xp.ndarray):
                        indices_temp = xp.expand_dims(xp.asarray(indices_temp), (-1,))

            # prepare slicing information for entry
            if isinstance(values, np.ndarray):
                inds0 = np.repeat(
                    np.arange(len(indices_temp))[:, None], indices_temp.shape[1], axis=1
                )
            elif isinstance(values, xp.ndarray):
                inds0 = xp.repeat(
                    np.arange(len(indices_temp))[:, None], indices_temp.shape[1], axis=1
                )
            # self.xp.put_along_axis(self.holder[name], indices_temp, values_in[name], axis)
            # because cupy does not have put_along_axis
            self.holder[name][(inds0.flatten(), indices_temp.flatten())] = values_in[
                name
            ].reshape((-1,) + values_in[name].shape[2:])

    @property
    def flat(self):
        """Get flattened arrays from the stored objects.

        Here "flat" is in relation to ``self.base_shape``. Beyond ``self.base_shape``, the shape is mainted.
        
        """
        out = {}
        # loop through holder
        for name, values in self.holder.items():
            if (
                isinstance(values, np.ndarray) and values.dtype.name != "object"
            ) or isinstance(values, xp.ndarray):
                # need to account for higher dimensional arrays.
                out[name] = values.reshape((-1,) + values.shape[2:])
            else:
                out[name] = values.flatten()
        return out


class Branch(object):
    """Special container for one branch (model)

    This class is a key component of Eryn. It this type of object
    that allows for different models to be considered simultaneously
    within an MCMC run.

    Args:
        coords (4D double np.ndarray[ntemps, nwalkers, nleaves_max, ndim]): The coordinates
            in parameter space of all walkers.
        inds (3D bool np.ndarray[ntemps, nwalkers, nleaves_max], optional): The information
            on which leaves are used and which are not used. A value of True means the specific leaf
            was used in this step. Parameters from unused walkers are still kept. When they
            are output to the backend, the backend saves a special number (default: ``np.nan``) for all coords
            related to unused leaves at that step. If None, inds will fill with all True values.
            (default: ``None``)
        branch_supplimental (object): :class:`BranchSupplimental` object specific to this branch. (default: ``None``)

    Raises:
        ValueError: ``inds`` has wrong shape or number of leaves is less than zero.

    """

    def __init__(self, coords, inds=None, branch_supplimental=None):

        # store branch info
        self.coords = coords
        self.ntemps, self.ntrees, self.nleaves_max, self.ndim = coords.shape
        self.shape = coords.shape

        # make sure inds is correct
        if inds is None:
            self.inds = np.full((self.ntemps, self.ntrees, self.nleaves_max), True)
        elif not isinstance(inds, np.ndarray):
            raise ValueError("inds must be np.ndarray in Branch.")
        elif inds.shape != (self.ntemps, self.ntrees, self.nleaves_max):
            raise ValueError("inds has wrong shape.")
        else:
            self.inds = inds

        if branch_supplimental is not None:
            # make sure branch_supplimental shape matches
            if branch_supplimental.base_shape != self.inds.shape:
                raise ValueError(
                    f"branch_supplimental shape ( {branch_supplimental.base_shape} ) does not match inds shape ( {self.inds.shape} )."
                )

        # store
        self.branch_supplimental = branch_supplimental

    @property
    def nleaves(self):
        """Number of leaves for each walker"""
        # get number of leaves in each walker by summing inds along last axis
        nleaves = np.sum(self.inds, axis=-1)
        return nleaves


class State(object):
    """The state of the ensemble during an MCMC run

    Args:
        coords (double ndarray[ntemps, nwalkers, nleaves_max, ndim], dict, or :class:`.State`): The current positions of the walkers
            in the parameter space. If dict, need to use ``branch_names`` for the keys.
        inds (bool ndarray[ntemps, nwalkers, nleaves_max] or dict, optional): The information
            on which leaves are used and which are not used. A value of True means the specific leaf
            was used in this step. If dict, need to use ``branch_names`` for the keys.
            Input should be ``None`` if a complete :class:`.State` object is input for ``coords``.
            (default: ``None``)
        log_like (ndarray[ntemps, nwalkers], optional): Log likelihoods
            for the  walkers at positions given by ``coords``.
            Input should be ``None`` if a complete :class:`.State` object is input for ``coords``.
            (default: ``None``)
        log_prior (ndarray[ntemps, nwalkers], optional): Log priors
            for the  walkers at positions given by ``coords``.
            Input should be ``None`` if a complete :class:`.State` object is input for ``coords``.
            (default: ``None``)
        betas (ndarray[ntemps], optional): Temperatures in the sampler at the current step.
            Input should be ``None`` if a complete :class:`.State` object is input for ``coords``.
            (default: ``None``)
        blobs (ndarray[ntemps, nwalkers, nblobs], Optional): The metadata “blobs”
            associated with the current position. The value is only returned if
            lnpostfn returns blobs too.
            Input should be ``None`` if a complete :class:`.State` object is input for ``coords``.
            (default: ``None``)
        random_state (Optional): The current state of the random number
            generator.
            Input should be ``None`` if a complete :class:`.State` object is input for ``coords``.
            (default: ``None``)
        copy (bool, optional): If True, copy the the arrays in the former :class:`.State` obhect.

    Raises:
        ValueError: Dimensions of inputs or input types are incorrect.

    """

    # __slots__ = (
    #    "branches",
    #    "log_like",
    #    "log_prior",
    #    "blobs",
    #    "betas",
    #    "supplimental",
    #    "random_state",
    # )

    def __init__(
        self,
        coords,
        inds=None,
        branch_supplimental=None,
        supplimental=None,
        log_like=None,
        log_prior=None,
        betas=None,
        blobs=None,
        random_state=None,
        copy=False,
    ):
        # decide if copying input info
        dc = deepcopy if copy else lambda x: x

        # check if coords is a State object
        if hasattr(coords, "branches"):
            self.branches = dc(coords.branches)
            self.log_like = dc(coords.log_like)
            self.log_prior = dc(coords.log_prior)
            self.blobs = dc(coords.blobs)
            self.betas = dc(coords.betas)
            self.supplimental = dc(coords.supplimental)
            self.random_state = dc(coords.random_state)
            return

        # protect against simplifying settings
        if isinstance(coords, np.ndarray):
            coords = {"model_0": coords}
        elif not isinstance(coords, dict):
            raise ValueError(
                "Input coords need to be np.ndarray, dict, or State object."
            )

        for name in coords:
            if coords[name].ndim == 2:
                coords[name] = coords[name][None, :, None, :]

            # assume (ntemps, nwalkers) provided
            if coords[name].ndim == 3:
                coords[name] = coords[name][:, :, None, :]

            elif coords[name].ndim < 2 or coords[name].ndim > 4:
                raise ValueError(
                    "Dimension off coordinates must be between 2 and 4. coords dimension is {0}.".format(
                        coords.ndim
                    )
                )

        # if no inds given, make sure this is clear for all Branch objects
        if inds is None:
            inds = {key: None for key in coords}
        elif not isinstance(inds, dict):
            raise ValueError("inds must be None or dict.")

        if branch_supplimental is None:
            branch_supplimental = {key: None for key in coords}
        elif not isinstance(branch_supplimental, dict):
            raise ValueError("branch_supplimental must be None or dict.")

        # setup all information for storage
        self.branches = {
            key: Branch(
                dc(temp_coords),
                inds=inds[key],
                branch_supplimental=branch_supplimental[key],
            )
            for key, temp_coords in coords.items()
        }
        self.log_like = dc(np.atleast_2d(log_like)) if log_like is not None else None
        self.log_prior = dc(np.atleast_2d(log_prior)) if log_prior is not None else None
        self.blobs = dc(np.atleast_3d(blobs)) if blobs is not None else None
        self.betas = dc(np.atleast_1d(betas)) if betas is not None else None
        self.supplimental = dc(supplimental)
        self.random_state = dc(random_state)

    @property
    def branches_inds(self):
        """Get the ``inds`` from all branch objects returned as a dictionary with ``branch_names`` as keys."""
        return {name: branch.inds for name, branch in self.branches.items()}

    @property
    def branches_coords(self):
        """Get the ``coords`` from all branch objects returned as a dictionary with ``branch_names`` as keys."""
        return {name: branch.coords for name, branch in self.branches.items()}

    @property
    def branches_supplimental(self):
        """Get the ``branch.supplimental`` from all branch objects returned as a dictionary with ``branch_names`` as keys."""
        return {
            name: branch.branch_supplimental for name, branch in self.branches.items()
        }

    @property
    def branch_names(self):
        """Get the branch names in this state."""
        return list(self.branches.keys())

    def copy_into_self(self, state_to_copy):
        for name in state_to_copy.__slots__:
            setattr(self, name, getattr(state_to_copy, name))

    def get_log_posterior(self, temper: bool = False):
        """Get the posterior probability
        
        Args:
            temper (bool, optional): If ``True``, apply tempering to the posterior computation.

        Returns:
            np.ndarray[ntemps, nwalkers]: Log of the posterior probability.
        
        """

        if temper:
            betas = self.betas

        else:
            betas = np.ones_like(self.betas)

        return betas * self.log_like + self.log_prior

    """
    # TODO
    def __repr__(self):
        return "State({0}, log_like={1}, blobs={2}, betas={3}, random_state={4})".format(
            self.coords, self.log_like, self.blobs, self.betas, self.random_state
        )

    def __iter__(self):
        temp = (self.coords,)
        if self.log_like is not None:
            temp += (self.log_like,)

        if self.blobs is not None:
            temp += (self.blobs,)

        if self.betas is None:
            temp += (self.betas,)

        if self.random_state is not None:
            temp += (self.random_state,)
        return iter(temp)
    """



# -*- coding: utf-8 -*-

from abc import ABC

import numpy as np


class Update(ABC, object):
    """Update the sampler."""

    @classmethod
    def __call__(self, iter, last_sample, sampler):
        """Call update function.

        Args:
            iter (int): Iteration of the sampler.
            last_sample (obj): Last state of sampler (:class:`eryn.state.State`).
            sampler (obj): Full sampler oject (:class:`eryn.ensemble.EnsembleSampler`).

        """
        raise NotImplementedError


class AdjustStretchProposalScale(Update):
    def __init__(
        self,
        target_acceptance=0.22,
        supression_factor=0.1,
        max_change=0.5,
        verbose=False,
    ):
        """Adjusted scale for stretch proposal based on cold chain acceptance rate"""
        self.target_acceptance = target_acceptance
        self.verbose = verbose
        self.max_change, self.supression_factor = max_change, supression_factor

        self.time = 0

    def __call__(self, iter, last_sample, sampler):

        mean_af = 0.0
        change = 1.0
        if self.time > 0:
            # cold chain -> 0
            mean_af = np.mean(
                (sampler.backend.accepted[:, 0] - self.previously_accepted)
                / (sampler.backend.iteration - self.previous_iter)
            )

            if mean_af > self.target_acceptance:
                factor = self.supression_factor * (mean_af / self.target_acceptance)
                if factor > self.max_change:
                    factor = self.max_change
                change = 1 + self.supression_factor * factor

            else:
                factor = self.supression_factor * (self.target_acceptance / mean_af)
                if factor > self.max_change:
                    factor = self.max_change
                change = 1 - factor

            sampler._moves[0].a *= change

        self.previously_accepted = sampler.backend.accepted[:, 0].copy()
        print(
            self.previously_accepted, "\n", mean_af, change, "\n", sampler._moves[0].a
        )
        self.previous_iter = sampler.backend.iteration
        self.time += 1



# from future.utils import iteritems
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

# setup version file
with open("README.md", "r") as fh:
    lines = fh.readlines()

for line in lines:
    if line.startswith("Current Version"):
        version_string = line.split("Current Version: ")[1].split("\n")[0]

with open("eryn/_version.py", "w") as f:
    f.write("__version__ = '{}'".format(version_string))

setup(
    name="Eryn",
    # Random metadata. there's more you can supply
    author_email="mikekatz04@gmail.com",
    author="Michael Katz, Nikos Karnesis",
    description="An all purpose MCMC sampler",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=version_string,
    url="https://github.com/mikekatz04/Eryn",
    packages=[
        "eryn",
        "eryn.backends",
        "eryn.moves",
        "eryn.utils",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.6",
)



# -*- coding: utf-8 -*-

from abc import ABC

import numpy as np


class Stopping(ABC, object):
    """Base class for stopping.
    
    Stopping checks are only performed every ``thin_by`` iterations.
    
    """

    @classmethod
    def __call__(self, iter, last_sample, sampler):
        """Call update function.

        Args:
            iter (int): Iteration of the sampler.
            last_sample (obj): Last state of sampler (:class:`eryn.state.State`).
            sampler (obj): Full sampler oject (:class:`eryn.ensemble.EnsembleSampler`).

        Returns:
            bool: Value of ``stop``. If ``True``, stop sampling.
            
        """
        raise NotImplementedError


class SearchConvergeStopping(Stopping):
    """Stopping function based on a convergence to a maximunm Likelihood.

    Stopping checks are only performed every ``thin_by`` iterations.
    Therefore, the iterations of stopping checks are really every
    ``sampler iterations * thin_by``.  

    All arguments are stored as attributes.

    Args:
        n_iters (int, optional): Number of iterative stopping checks that need to pass
            in order to stop the sampler. (default: ``30``)
        diff (float, optional): Change in the Likelihood needed to fail the stopping check. In other words,
            if the new maximum Likelihood is more than ``diff`` greater than the old, all iterative checks 
            reset. (default: 0.1). 
        start_iteration (int, optional): Iteration of sampler to start checking to stop. (default: 0)
        verbose (bool, optional): If ``True``, print information. (default: ``False``)

    Attributes:
        iters_consecutive (int): Number of consecutive passes of the stopping check.
        past_like_best (float): Previous best Likelihood. The initial value is ``-np.inf``.
    
    """

    def __init__(self, n_iters=30, diff=0.1, start_iteration=0, verbose=False):

        # store all the relevant information
        self.n_iters = n_iters

        self.diff = diff
        self.verbose = verbose
        self.start_iteration = start_iteration

        # initialize important info
        self.iters_consecutive = 0
        self.past_like_best = -np.inf

    def __call__(self, iter, sample, sampler):
        """Call update function.

        Args:
            iter (int): Iteration of the sampler.
            last_sample (obj): Last state of sampler (:class:`eryn.state.State`).
            sampler (obj): Full sampler oject (:class:`eryn.ensemble.EnsembleSampler`).

        Returns:
            bool: Value of ``stop``. If ``True``, stop sampling.
            
        """

        # if we have not reached the start iteration return
        if iter < self.start_iteration:
            return False

        # get best Likelihood so far
        like_best = sampler.get_log_like(discard=self.start_iteration).max()

        # compare to last
        # if it is less than diff change it passes
        if np.abs(like_best - self.past_like_best) < self.diff:
            self.iters_consecutive += 1

        else:
            # if it fails reset iters consecutive
            self.iters_consecutive = 0

            # store new best
            self.past_like_best = like_best

        # print information
        if self.verbose:
            print(
                f"\nITERS CONSECUTIVE: {self.iters_consecutive}",
                f"Previous best LL: {self.past_like_best}",
                f"Current best LL: {like_best}\n",
            )

        if self.iters_consecutive >= self.n_iters:
            # if we have passes the number of iters necessary, return True and reset
            self.iters_consecutive = 0
            return True

        else:
            return False


"""
class AutoCorrelationStop(Stopping):
    # TODO: check and doc this
    def __init__(self, autocorr_multiplier=50, verbose=False):
        self.autocorr_multiplier = autocorr_multiplier
        self.verbose = verbose

        self.time = 0

    def __call__(self, iter, last_sample, sampler):

        tau = sampler.backend.get_autocorr_time(multiply_thin=False)

        if self.time > 0:
            # backend iteration
            iteration = sampler.backend.iteration

            finish = []

            for name, values in tau.items():
                converged = np.all(tau[name] * self.autocorr_multiplier < iteration)
                converged &= np.all(
                    np.abs(self.old_tau[name] - tau[name]) / tau[name] < 0.01
                )

                finish.append(converged)

            stop = True if np.all(finish) else False
            if self.verbose:
                print(
                    "\ntau:",
                    tau,
                    "\nIteration:",
                    iteration,
                    "\nAutocorrelation multiplier:",
                    self.autocorr_multiplier,
                    "\nStopping:",
                    stop,
                    "\n",
                )

        else:
            stop = False

        self.old_tau = tau
        self.time += 1
        return stop
"""
